/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "trendview.hh"
#include "jsonStack.hh"
#include "FilterDesign.hh"
#include "cubic_interpolate.hh"
#include "ReadTrend.hh"
#include <stdexcept>

using namespace std;

EXECDAT(trendview)

//======================================  Channel info class constructor
trendview::channel_info::channel_info(const std::string& name)
  : mDataChannel(name), mTrendChannel(name), mViewChannel(name),
   mSettle(0.0), mViewEpoch(43200), mViewStride(15.0), mDebug(0),
   mStartTime(0), mCurrent(0), mAccTime(0)
{
   set_view_epoch(mViewEpoch);
}

//======================================  Channel info class destructor
trendview::channel_info::~channel_info(void) {
}

//======================================  Channel info class destructor
void 
trendview::channel_info::add_data(const TSeries& ts) {
   Time tIndex = ts.getStartTime();
   if (debug()) cout << "Process " << mDataChannel << " data at "
		     << tIndex.getS() << ":" << ts.getInterval() << endl;
   
   //-----------------------------------  First entry
   if (!mCurrent || mCurrent != tIndex) {
      if (!mFilter.empty()) {
	 double srate = 1.0/ts.getTStep();
	 FilterDesign fd(mFilter.c_str(), srate);
	 mPipe.set(fd.release());
      }
      reset(tIndex);
   }

   //------------------------------------  Optional filtering
   if (!mPipe.null()) {
      mFiltered = mPipe(ts);
      if (tIndex < mStartTime + mSettle) {
	 mFiltered.eraseStart(mStartTime + mSettle - tIndex);
      }
   }
   else {
      mFiltered = ts;
   }

   //-------------------------------------  Set the accumulator start time
   Time fIndex = mFiltered.getStartTime();   
   if (!mAccTime) {
      long dt = fIndex.getS() % long(mViewStride);
      mAccTime = Time(fIndex.getS() - dt);
   }

   //-------------------------------------  Loop over filtered data strides 
   Time fStop  = mFiltered.getEndTime();
   while (fIndex < fStop) {
      Time vStop = mAccTime + mViewStride; 
      Interval left = fStop - fIndex;
      if (vStop < fStop) left = vStop - fIndex;
      mViewAcc.addData(mFiltered.extract(fIndex, left));
      if (fIndex + left >= vStop) {
	 if (debug()) cout << "append " << mViewChannel << " view for "
			   << mAccTime.getS() << ":" << mViewStride << endl;
	 float xmean =  mViewAcc.getAverage();
	 mData.fixedAppend(mAccTime, mViewStride, &xmean);
	 float xmin  =  mViewAcc.getMinimum();	 
	 mMinimum.fixedAppend(mAccTime, mViewStride, &xmin);
	 float xmax  =  mViewAcc.getMaximum();
	 mMaximum.fixedAppend(mAccTime, mViewStride, &xmax);
	 float xsigma =  mViewAcc.getSigma();
	 mSigma.fixedAppend(mAccTime, mViewStride, &xsigma);
	 mViewAcc.reset();
	 mAccTime = vStop;
      }
      fIndex += left;
   }
   mCurrent = ts.getEndTime();
}

//======================================  Channel info class destructor
void 
trendview::channel_info::book_trend(Trend& trender) {
   trender.addChannel(mTrendChannel.c_str());
}

//======================================  book channels in the monitor server
void 
trendview::channel_info::book_viewer(MonServer& viewer) const {
   viewer.serveData((mViewChannel+"_mean").c_str(),
		    static_cast<const TSeries*>(&mData));
   viewer.serveData((mViewChannel+"_min").c_str(),
		    static_cast<const TSeries*>(&mMinimum));
   viewer.serveData((mViewChannel+"_max").c_str(),
		    static_cast<const TSeries*>(&mMaximum));
   viewer.serveData((mViewChannel+"_sigma").c_str(),
		    static_cast<const TSeries*>(&mSigma));
}

//======================================  book channels in the monitor server
void 
trendview::channel_info::display(std::ostream& out) const {
   out << "data channel:   " << mDataChannel << endl;
   out << "trend channel:  " << mTrendChannel << endl;
   out << "viewer channel: " << mViewChannel << endl;
   out << "filter design:  " << mFilter << endl;
   out << "settle time:    " << mSettle << endl;
   out << "Viewer epoch:   " << mViewEpoch << endl;
   out << "viewer stride:  " << mViewStride << endl;
   out << "debug level:    " << mDebug << endl;
}

//======================================  Initialize viewer histories
void
trendview::channel_info::init_data(const std::string& monitor) {
   ReadTrend::string_vect rt_name;
   ReadTrend::type_vect rt_type;
   rt_name.push_back(mTrendChannel); rt_type.push_back(ReadTrend::kMean);
   rt_name.push_back(mTrendChannel); rt_type.push_back(ReadTrend::kMin);
   rt_name.push_back(mTrendChannel); rt_type.push_back(ReadTrend::kMax);
   rt_name.push_back(mTrendChannel); rt_type.push_back(ReadTrend::kSigma);
   ReadTrend::tseries_vect rt_data(4);

   ReadTrend rt;
   Time history_start = Now() - mViewEpoch;
   // rt.setDebug();
   rt.setType(ReadTrend::kMinute);
   rt.setMonitor(monitor.c_str());
   Interval dT = mViewEpoch;
   rt.getSeries(rt_name, rt_type, history_start, mViewEpoch, rt_data);
   Time t0 = rt_data[0].getStartTime();
   for (int id=0; id<4; id++) {
      cubic_interpolate ci(int(60.0 / mViewStride));
      TSeries data = ci(rt_data[id]);
      switch (id) {
      case 0:
	 static_cast<TSeries&>(mData) = data;
	 break;
      case 1:
	 static_cast<TSeries&>(mMinimum) = data;
	 break;
      case 2:
	 static_cast<TSeries&>(mMaximum) = data;
	 break;
      case 3:
	 static_cast<TSeries&>(mSigma) = data;
	 break;
      }
   }

}

//======================================  get last stride of filtered data.
void
trendview::channel_info::reset(const Time& tIndex) {
   mCurrent   = tIndex;
   mStartTime = tIndex;
   mAccTime   = Time(0);
   mPipe.reset();
   mViewAcc.reset();
   mFiltered.Clear();
}
      
//======================================  Channel info class destructor
void 
trendview::channel_info::set_debug(int dbg) {
   mDebug = dbg;
}
      
//======================================  Channel info class destructor
void 
trendview::channel_info::set_filter(const std::string& filt) {
   mFilter = filt;
}
       
//======================================  Channel info class destructor
void 
trendview::channel_info::set_trend_channel(const std::string& tchan) {
   mTrendChannel = tchan;
}

//======================================  Channel info class destructor
void 
trendview::channel_info::set_view_channel(const std::string& vchan) {
   mViewChannel = vchan;
}


//======================================  Channel info class destructor
void 
trendview::channel_info::set_view_epoch(Interval epoch) {
   mViewEpoch = epoch;
   mData.setMaxLen(epoch);
   mMinimum.setMaxLen(epoch);
   mMaximum.setMaxLen(epoch);
   mSigma.setMaxLen(epoch);
}

//======================================  Channel info class destructor
void 
trendview::channel_info::set_view_stride(Interval vstride) {
   mViewStride = vstride;
}

//======================================  Trendview constructor.
trendview::trendview(int argc, const char *argv[])
   : DatEnv(argc, argv)
{
   string config = "trendview.json";
   for (int i=1; i<argc; i++) {
      string argi = argv[i];
      if (argi == "-conf") {
	 config = argv[++i];
      }
      else if (isDatEnvArg(argv[i])) {
	 i++;
      }
      else {
	 cerr << "Unrecognized argument: " << argi << endl;
	 finish();
	 return;
      }
   }
   
   //-----------------------------------  Configure the monitor
   configure(config);

   //-----------------------------------  Set up server
   if (!mServerName.empty()) {
      MonServer::setServerName(mServerName.c_str());
   }

   //-----------------------------------  Set up trender
   if (!mTrendName.empty()) {
      mTrend.setType(Trend::kMinute);
      mTrend.setName(mTrendName.c_str());
   }
   
   //-----------------------------------  Set up channels
   if (Debug()) {
      cout << "Configured channels: " << endl
	   << "-------------------- " << endl;
   }
   size_t N = mChannel.size();
   for (size_t i=0; i<N; i++) {
      channel_info& chan_i(mChannel[i]);
      if (Debug()) {
	 chan_i.display(cout);
      }
      getDacc().addChannel(chan_i.data_channel());
      chan_i.book_trend(mTrend);
      chan_i.book_viewer(static_cast<MonServer&>(*this));
      chan_i.init_data(mTrendName);
   }

}

//======================================  Trendview destructor.
trendview::~trendview(void) {
}

//======================================  Configure trendview monitor.
void
trendview::configure(const std::string& file) {
   jsonStack jstk(file);
   //-----------------------------------  Get the server name
   jstk.fetch_data("server-name", mServerName);
   jstk.fetch_data("trend-name",  mTrendName);
   
   if (jstk.push_member("channels")) {
      jsonStack::autopop pop_array(jstk);
      while(jstk.iterate()) {
	 jsonStack::autopop pop_iter(jstk);
	 string channel;
	 if (!jstk.fetch_data("channel-name", channel)) {
	    throw runtime_error("trendview::configure: Unnamed channel");
	 }
	 mChannel.push_back(channel_info(channel));
	 if (Debug()) cout << "channel name: " << channel << endl;

	 mChannel.back().set_debug(Debug());
	 string tmpstr;
	 if (jstk.fetch_data("viewer-name", tmpstr)) {
	    mChannel.back().set_view_channel(tmpstr);
	 }
	 if (jstk.fetch_data("trend-name", tmpstr)) {
	    mChannel.back().set_trend_channel(tmpstr);
	 }
	 if (jstk.fetch_data("filter-design", tmpstr)) {
	    mChannel.back().set_filter(tmpstr);
	 }
	 double tmpdbl = 0;
	 if (jstk.fetch_data("viewer-epoch", tmpstr)) {
	    mChannel.back().set_view_epoch(tmpdbl);
	 }
	 if (jstk.fetch_data("viewer-stride", tmpstr)) {
	    mChannel.back().set_view_stride(tmpdbl);
	 }
      }
   }
   if (mChannel.empty()) {
      throw runtime_error("trendview: No channels defined");
   }
}

//======================================  Process data
void
trendview::ProcessData(void) {
   if (Debug()) cout << "Processing stride at " << getDacc().getFillTime()
		     << endl;
   //-----------------------------------  Set up chnnels
   size_t N = mChannel.size();
   for (size_t i=0; i<N; i++) {
      channel_info& chan_i(mChannel[i]);
      string  data_chan = chan_i.data_channel();
      const TSeries* raw = getDacc().refData(data_chan);
      if (!raw || raw->isEmpty()) continue;
      chan_i.add_data(*raw);
      const TSeries& filtered = chan_i.filtered_data();
      string trend_chan = chan_i.trend_channel();
      if (trend_chan.empty()) continue;
      mTrend.trendData(trend_chan.c_str(), filtered);
   }
   if (!mTrendName.empty()) {
      mTrend.Update(getDacc().getCurrentTime());
   }
}

//======================================  Attention handler
void
trendview::Attention(void) {
   MonServer::Attention();
}
